# pylint: disable=too-few-public-methods
# pylint: disable=invalid-name
# pylint: disable=no-self-use
# pylint: disable=unused-argument

from pathlib import Path
import os
import random
import string
import pytest

from pyhstr import (
    __main__,
    application,
    utilities,
    user_interface,
)
from pyhstr.utilities import Shell, read, write


class FakeReadline:
    def __init__(self):
        self.history = read("tests/history/fake_python_history")

    def get_history_item(self, i):
        return self.history[i - 1]

    def remove_history_item(self, i):
        self.history.pop(i)

    def write_history_file(self, path):
        write(path, self.history)

    def get_current_history_length(self):
        return len(self.history)


class FakeCurses:
    LINES = 10
    COLS = 80

    class error(Exception):
        pass


class FakeStdscr:
    def __init__(self):
        self.addstred = []
        self.addched = []

    def addstr(self, *args):
        self.addstred.append(args)

    def addch(self, *args):
        self.addched.append(args)

    def box(self, *args):
        pass

    def timeout(self, *args):
        pass

    def subwin(self, *args):
        return FakeStdscr()

    def clear(self):
        pass

    def erase(self):
        pass

    def refresh(self):
        pass

    def getch(self):
        pass

    def nodelay(self, *args):
        pass

    def attron(self, *args):
        pass

    def attroff(self, *args):
        pass

    def getmaxyx(self):
        return FakeCurses.LINES, FakeCurses.COLS


@pytest.fixture
def fake_readline(monkeypatch):
    monkeypatch.setattr(application, "readline", FakeReadline())


@pytest.fixture
def fake_curses(monkeypatch):
    fc = FakeCurses()
    monkeypatch.setattr(__main__, "curses", fc)
    monkeypatch.setattr(user_interface, "curses", fc)


@pytest.fixture
def fake_stdscr():
    return FakeStdscr()


class FakeTermios:
    TIOCSTI = None


class FakeFcntl:
    def __init__(self):
        self.echoed = []

    def ioctl(self, *args):
        self.echoed.append(args)


@pytest.fixture
def fake_fcntl(monkeypatch):
    fake_f = FakeFcntl()
    monkeypatch.setattr(utilities, "fcntl", fake_f)
    return fake_f


@pytest.fixture
def fake_termios(monkeypatch):
    monkeypatch.setattr(utilities, "termios", FakeTermios())


def fake_get_ipython_history():
    return read(application.SHELLS[Shell.IPYTHON]["hist"])


@pytest.fixture
def fake_ipython(monkeypatch):
    monkeypatch.setattr(application, "SHELL", Shell.IPYTHON)
    monkeypatch.setattr(application, "get_ipython_history", fake_get_ipython_history)
    monkeypatch.setitem(
        application.SHELLS,
        Shell.IPYTHON,
        {
            "hist": Path("tests/history/fake_ipython_history"),
            "fav": Path("tests/favorites/fake_ipython_favorites"),
        },
    )


@pytest.fixture
def fake_bpython(monkeypatch):
    monkeypatch.setattr(application, "SHELL", Shell.BPYTHON)
    monkeypatch.setattr(help, "__module__", "bpython")
    monkeypatch.setitem(
        application.SHELLS,
        Shell.BPYTHON,
        {
            "hist": Path("tests/history/fake_bpython_history"),
            "fav": Path("tests/favorites/fake_bpython_favorites"),
        },
    )


@pytest.fixture
def fake_standard(monkeypatch):
    monkeypatch.setattr(application, "SHELL", Shell.STANDARD)
    monkeypatch.setitem(
        application.SHELLS,
        Shell.STANDARD,
        {
            "hist": Path("tests/history/fake_python_history"),
            "fav": Path("tests/favorites/fake_python_favorites"),
        },
    )


@pytest.fixture
def random_history(request):
    length = request.node.get_closest_marker("history_length").args[0]
    chars = string.ascii_letters + string.digits + r"!@#$%^&*()_+,./;'\\[]<>?:\"|{}"
    return [
        "".join([random.choice(chars) for _ in range(random.randint(40, 80))])
        for _ in range(length)
    ]


params = [
    pytest.param(
        Shell.STANDARD,
        pytest.lazy_fixture("fake_standard"),
        marks=[
            pytest.mark.skipif(
                os.environ["PYTHON_SHELL"] != "python",
                reason="standard only",
            ),
            pytest.mark.python,
        ],
    ),
    pytest.param(
        Shell.BPYTHON,
        pytest.lazy_fixture("fake_bpython"),
        marks=[
            pytest.mark.skipif(
                os.environ["PYTHON_SHELL"] != "bpython",
                reason="bpython only",
            ),
            pytest.mark.bpython,
        ],
    ),
    pytest.param(
        Shell.IPYTHON,
        pytest.lazy_fixture("fake_ipython"),
        marks=[
            pytest.mark.skipif(
                os.environ["PYTHON_SHELL"] != "ipython",
                reason="ipython only",
            ),
            pytest.mark.ipython,
        ],
    ),
]



import curses
from typing import Any, TYPE_CHECKING

from pyhstr.application import App, View, SHELLS, SHELL
from pyhstr.utilities import echo, write

if TYPE_CHECKING:
    from _curses import _CursesWindow  # pylint: disable=no-name-in-module
else:
    _CursesWindow = Any


KEY_BINDINGS = {
    curses.KEY_UP: -1,
    curses.KEY_DOWN: 1,
    curses.KEY_PPAGE: -1,
    curses.KEY_NPAGE: 1,
}

CTRL_E = "\x05"
CTRL_F = "\x06"
CTRL_T = "\x14"
CTRL_SLASH = "\x1f"
TAB = "\t"
ENTER = "\n"
ESC = "\x1b"
DEL = curses.KEY_DC


def main(stdscr: _CursesWindow) -> None:  # pylint: disable=too-many-statements
    app = App(stdscr)
    app.user_interface.init_color_pairs()
    app.user_interface.populate_screen()

    while True:
        try:
            user_input = app.stdscr.get_wch()
        except curses.error:
            continue
        except KeyboardInterrupt:
            break

        # user_input is Union[int, str], sometimes isinstance needed to make mypy happy

        if user_input == CTRL_E:
            app.toggle_regex_mode()
            app.user_interface.page.selected.value = 0
            app.user_interface.populate_screen()

        elif user_input == CTRL_F:
            command = app.user_interface.page.get_selected()
            if app.view == View.FAVORITES:
                page_size = app.user_interface.page.get_size() - 1
                selected = app.user_interface.page.selected
                if selected.value == page_size:
                    selected.move(-1)
            app.add_or_rm_fav(command)
            write(SHELLS[SHELL]["fav"], app.commands[View.FAVORITES])
            app.stdscr.clear()
            app.user_interface.populate_screen()

        elif user_input == TAB:
            command = app.user_interface.page.get_selected()
            echo(command)
            break

        elif user_input == ENTER:
            command = app.user_interface.page.get_selected()
            echo(command)
            echo("\n")
            break

        elif user_input == CTRL_T:
            app.toggle_case()
            app.user_interface.populate_screen()

        elif user_input == ESC:
            break

        elif user_input == CTRL_SLASH:
            app.toggle_view()
            app.user_interface.page.selected.value = 0
            app.user_interface.page.value = 1
            app.stdscr.clear()
            app.user_interface.populate_screen()

        elif user_input in {curses.KEY_UP, curses.KEY_DOWN}:
            assert isinstance(user_input, int)
            app.user_interface.page.selected.move(KEY_BINDINGS[user_input])
            app.user_interface.populate_screen()

        elif user_input in {curses.KEY_NPAGE, curses.KEY_PPAGE}:
            assert isinstance(user_input, int)
            app.user_interface.page.turn(KEY_BINDINGS[user_input])
            app.user_interface.populate_screen()

        elif user_input == curses.KEY_BACKSPACE:
            app.search_string = app.search_string[:-1]
            if not app.search_string:
                app.user_interface.page.selected.value = 0
            app.commands = app.to_restore.copy()
            app.search()

        elif user_input == DEL:
            command = app.user_interface.page.get_selected()
            app.user_interface.prompt_for_deletion(command)
            answer = app.stdscr.getch()
            if answer == ord("y"):
                page_size = app.user_interface.page.get_size() - 1
                selected = app.user_interface.page.selected
                if selected.value == page_size:
                    selected.move(-1)
                app.delete_from_history(command)
            app.stdscr.clear()
            app.user_interface.populate_screen()


        elif isinstance(user_input, str):
            # not another special int character like curses.KEY_UP
            app.search_string += user_input
            app.commands = app.to_restore.copy()
            app.search()

    stdscr.clear()
    stdscr.refresh()
    curses.doupdate()



import collections
import enum
import fcntl
import termios
from pathlib import Path
from typing import List, Optional

try:
    import IPython
except ModuleNotFoundError:  # pragma: no cover
    IPython = None

try:
    from bpython.config import (
        Struct,
        get_config_home,
        loadini,
    )
except ModuleNotFoundError:  # pragma: no cover
    Struct = get_config_home = loadini = None


class Shell(enum.Enum):
    STANDARD = "python"
    IPYTHON = "ipython"
    BPYTHON = "bpython"


class View(enum.Enum):
    SORTED = 0
    FAVORITES = 1
    ALL = 2


def sort(thing: List[str]) -> List[str]:
    return remove_duplicates(
        sorted(
            sorted(
                thing,
                key=lambda item: {x: i for i, x in enumerate(thing)}[item],
                reverse=True,
            ),
            key=lambda item: collections.Counter(thing)[item],
            reverse=True,
        )
    )


def write(path: Optional[Path], thing: List[str]) -> None:
    assert path is not None
    with open(path, "w") as f:
        for thingy in thing:
            print(thingy, file=f)


def read(path: Optional[Path]) -> List[str]:
    assert path is not None
    try:
        with open(path, "r") as f:
            return [command.strip() for command in f]
    except FileNotFoundError:
        path.parent.mkdir(exist_ok=True)
        path.touch()
        return [""]


def echo(command: str) -> None:
    for byte in command.encode("utf-8"):
        fcntl.ioctl(0, termios.TIOCSTI, bytes([byte]))


def remove_duplicates(thing: List[str]) -> List[str]:
    return list(collections.OrderedDict.fromkeys(thing))


def get_ipython_history() -> List[str]:
    return [
        entry
        for session_number, line_number, entry in IPython.get_ipython().history_manager.search()
    ]


def get_bpython_history_path() -> Optional[Path]:
    if any(x is None for x in {Struct, get_config_home, loadini}):
        return None
    config = Struct()
    loadini(config, Path(get_config_home()).expanduser() / "config")
    return Path(config.hist_file).expanduser()


def is_ipython() -> bool:
    if IPython is not None:
        return IPython.get_ipython() is not None
    return False


def is_bpython() -> bool:
    return help.__module__.startswith("bpython")


def detect_shell() -> Shell:
    if is_ipython():
        return Shell.IPYTHON
    elif is_bpython():
        return Shell.BPYTHON
    return Shell.STANDARD



import re
import readline

from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, TYPE_CHECKING

try:
    import IPython
except ModuleNotFoundError:  # pragma: no cover
    IPython = None

from pyhstr.user_interface import UserInterface
from pyhstr.utilities import (
    Shell,
    View,
    detect_shell,
    get_bpython_history_path,
    get_ipython_history,
    read,
    remove_duplicates,
    sort,
    write,
)

if TYPE_CHECKING:
    from _curses import _CursesWindow  # pylint: disable=no-name-in-module
else:
    _CursesWindow = Any


SHELL = detect_shell()

SHELLS: Dict[Shell, Dict[str, Optional[Path]]] = {
    Shell.IPYTHON: {
        "hist": None,
        "fav": Path("~/.config/pyhstr/ipython_favorites").expanduser(),
    },
    Shell.BPYTHON: {
        "hist": get_bpython_history_path(),
        "fav": Path("~/.config/pyhstr/bpython_favorites").expanduser(),
    },
    Shell.STANDARD: {
        "hist": Path("~/.python_history").expanduser(),
        "fav": Path("~/.config/pyhstr/python_favorites").expanduser(),
    },
}


class App:
    def __init__(self, stdscr: _CursesWindow):
        self.stdscr = stdscr
        self.user_interface = UserInterface(self)
        self.raw_history: List[str] = self.get_history()
        self.commands: Dict[View, List[str]] = {
            View.SORTED: sort(self.raw_history),
            View.FAVORITES: sort(read(SHELLS[SHELL]["fav"])),
            View.ALL: remove_duplicates(self.raw_history),
        }
        self.to_restore = self.commands.copy()
        self.regex_mode: bool = False
        self.case_sensitivity: bool = False
        self.view: View = View.SORTED
        self.search_string = ""

    def get_history(self) -> List[str]:  # pylint: disable=no-self-use
        if SHELL == Shell.IPYTHON:
            return get_ipython_history()
        return read(SHELLS[SHELL]["hist"])

    def search(self) -> None:
        self.user_interface.page.selected.value = 0
        self.user_interface.page.value = 1

        search_regex = self.create_search_regex()

        if search_regex is not None:
            self.commands[self.view] = [
                cmd for cmd in self.commands[self.view] if search_regex.search(cmd)
            ]
            self.stdscr.clear()
            self.user_interface.populate_screen()
        else:
            self.commands[self.view] = []
            self.stdscr.clear()
            self.user_interface.populate_screen()
            self.user_interface.show_regex_error()

    def create_search_regex(self) -> Optional[Pattern]:
        try:
            search_string = (
                self.search_string if self.regex_mode else re.escape(self.search_string)
            )
            return re.compile(
                search_string, re.IGNORECASE if not self.case_sensitivity else 0
            )
        except re.error:
            return None

    def delete_python_history(self, command: str) -> None:  # pylint: disable=no-self-use
        readline_history = [
            readline.get_history_item(i + 1)
            for i in range(readline.get_current_history_length())
        ]

        cmd_indexes = [i for i, cmd in enumerate(readline_history) if cmd == command]

        for cmd_idx in reversed(cmd_indexes):
            readline.remove_history_item(cmd_idx)

        readline.write_history_file(str(SHELLS[Shell.STANDARD]["hist"]))

    def delete_ipython_history(self, command: str) -> None:  # pylint: disable=no-self-use
        IPython.get_ipython().history_manager.db.execute(
            "DELETE FROM history WHERE source=(?)", (command,)
        )

    def delete_bpython_history(self, command: str) -> None:
        self.raw_history = [cmd for cmd in self.raw_history if cmd != command]
        write(SHELLS[Shell.BPYTHON]["hist"], self.raw_history)

    def delete_from_history(self, command: str) -> None:
        if SHELL == Shell.STANDARD:
            self.delete_python_history(command)
        elif SHELL == Shell.IPYTHON:
            self.delete_ipython_history(command)
        elif SHELL == Shell.BPYTHON:
            self.delete_bpython_history(command)
        else:
            pass  # future implementations

        for view in self.commands.values():
            for cmd in view:
                if cmd == command:
                    view.remove(cmd)

        self.to_restore = self.commands.copy()

    def add_or_rm_fav(self, command: str) -> None:
        if command not in self.commands[View.FAVORITES]:
            self.commands[View.FAVORITES].append(command)
        else:
            self.commands[View.FAVORITES].remove(command)

    def toggle_regex_mode(self) -> None:
        self.regex_mode = not self.regex_mode

    def toggle_case(self) -> None:
        self.case_sensitivity = not self.case_sensitivity

    def toggle_view(self) -> None:
        self.view = View((self.view.value + 1) % 3)



"""History manager extension for the standard Python shell and IPython"""

__version__ = "0.5.1"

import curses
import sys

try:
    import IPython
except ModuleNotFoundError:  # pragma: no cover
    IPython = None

from pyhstr.__main__ import main
from pyhstr.application import SHELL
from pyhstr.utilities import Shell

hh = object()
original = sys.displayhook


def spam(arg: object) -> None:
    if arg == hh:
        curses.wrapper(main)
    else:
        original(arg)


if SHELL != Shell.IPYTHON:
    sys.displayhook = spam
else:

    @IPython.core.magic.register_line_magic
    def hh(line: str) -> None:  # pylint: disable=function-redefined,unused-argument
        """
        This line magic mirrors the behaviour of sys.displayhook
        in the regular Python shell. Use %hh to invoke.
        """
        curses.wrapper(main)



import curses
import sys
from typing import Dict, List, Union, TYPE_CHECKING
from pyhstr.utilities import View

if TYPE_CHECKING:
    from pyhstr.application import App  # pylint: disable=cyclic-import


COLORS: Dict[str, int] = {
    # yet to be initialized
    "normal": 0,
    "highlighted-white": 0,
    "highlighted-green": 0,
    "highlighted-red": 0,
    "white": 0,
    "bold-red": 0,
}

PYHSTR_LABEL = (
    "Type to filter, UP/DOWN move, RET/TAB select, DEL remove, ESC quit, C-f add/rm fav"
)
PYHSTR_STATUS = "- view:{} (C-/) - regex:{} (C-e) - case:{} (C-t) - page {}/{} -"

PS1 = getattr(sys, "ps1", ">>> ")

DISPLAY: Dict[str, Dict[Union[View, bool], str]] = {
    "view": {
        View.SORTED: "sorted",
        View.FAVORITES: "favorites",
        View.ALL: "history",
    },
    "case": {
        True: "sensitive",
        False: "insensitive",
    },
    "regex_mode": {
        True: "on",
        False: "off",
    },
}


class UserInterface:
    def __init__(self, app: "App"):
        self.app = app
        self.page = Page(self.app)

    def _addstr(self, y_coord: int, x_coord: int, text: str, color_info: int) -> None:
        """
        Works around curses' limitation of drawing at bottom right corner
        of the screen, as seen on https://stackoverflow.com/q/36387625
        """
        screen_height: int
        screen_width: int
        screen_height, screen_width = self.app.stdscr.getmaxyx()
        if x_coord + len(text) == screen_width and y_coord == screen_height - 1:
            try:
                self.app.stdscr.addstr(y_coord, x_coord, text, color_info)
            except curses.error:
                pass
        else:
            self.app.stdscr.addstr(y_coord, x_coord, text, color_info)

    @staticmethod
    def init_color_pairs() -> None:
        mapping: Dict[int, List[int]] = {
            1: [curses.COLOR_WHITE, curses.COLOR_BLACK],
            2: [curses.COLOR_BLACK, curses.COLOR_WHITE],
            3: [curses.COLOR_WHITE, curses.COLOR_GREEN],
            4: [curses.COLOR_WHITE, curses.COLOR_RED],
            5: [curses.COLOR_CYAN, curses.COLOR_BLACK],
            6: [curses.COLOR_RED, curses.COLOR_BLACK, curses.A_BOLD],
        }

        for idx, color in enumerate(COLORS, 1):
            curses.init_pair(idx, mapping[idx][0], mapping[idx][1])
            if len(mapping[idx]) > 2:
                COLORS[color] = curses.color_pair(idx) | mapping[idx][2]
            else:
                COLORS[color] = curses.color_pair(idx)

    def populate_screen(self) -> None:
        current_page = self.app.user_interface.page.value
        total_pages = self.app.user_interface.page.total_pages()

        status = PYHSTR_STATUS.format(
            DISPLAY["view"][self.app.view],
            DISPLAY["regex_mode"][self.app.regex_mode],
            DISPLAY["case"][self.app.case_sensitivity],
            current_page if total_pages > 0 else 0,
            total_pages,
        ).ljust(curses.COLS - 1)

        cmds = self.page.get_commands()

        for cmd_idx, cmd in enumerate(cmds):
            # print everything first (normal),
            # then print found matches (in red)
            # then print favorites (white)
            # then print selected on top of all that (green)
            try:
                padded_cmd = cmd.ljust(curses.COLS - 1)
                self._addstr(cmd_idx + 3, 1, padded_cmd, COLORS["normal"])
                matched_chars = self.get_matched_chars(cmd)

                if matched_chars:
                    for char_idx, char in enumerate(cmd):
                        if char_idx in matched_chars:
                            self.app.stdscr.attron(COLORS["bold-red"])
                            self.app.stdscr.addch(cmd_idx + 3, char_idx + 1, char)
                            self.app.stdscr.attroff(COLORS["bold-red"])

                if cmd in self.app.commands[View.FAVORITES]:
                    self._addstr(cmd_idx + 3, 1, padded_cmd, COLORS["white"])

                if cmd_idx == self.app.user_interface.page.selected.value:
                    self._addstr(
                        cmd_idx + 3, 1, padded_cmd, COLORS["highlighted-green"]
                    )
            except curses.error:
                pass

        self._addstr(1, 1, PYHSTR_LABEL, COLORS["normal"])
        self._addstr(2, 1, status, COLORS["highlighted-white"])
        self._addstr(0, 1, PS1 + self.app.search_string, COLORS["normal"])

    def prompt_for_deletion(self, command: str) -> None:
        prompt = f"Do you want to delete all occurences of {command}? y/n"
        self._addstr(1, 0, "".ljust(curses.COLS), COLORS["normal"])
        self._addstr(1, 1, prompt, COLORS["highlighted-red"])

    def show_regex_error(self) -> None:
        prompt = "Invalid regex. Try again."
        self._addstr(1, 0, "".ljust(curses.COLS), COLORS["normal"])
        self._addstr(1, 1, prompt, COLORS["highlighted-red"])
        self._addstr(0, 1, PS1 + self.app.search_string, COLORS["normal"])

    def get_matched_chars(self, command: str) -> List[int]:
        regex = self.app.create_search_regex()
        return (
            []
            if regex is None
            else [
                cmd_idx
                for m in regex.finditer(command)
                for cmd_idx in range(m.start(), m.end())
            ]
        )


class Page:
    def __init__(self, app: "App"):
        self.value = 1
        self.app = app
        self.selected = SelectedCmd(self)

    def turn(self, direction: int) -> None:
        """
        Paging starts from 1 but we want it to start at 0,
        because that's how our calculation with modulo works.

        So, if the cmd_idxing started from zero, we would have had:

        self.value = (self.value + 1) % total_pages

        ...which is increment and wrap around.

        Since we want the value to start at 1, we should:

        - subtract 1 from it when using it, because we want it to
        comply with the condition that page values start from 1,
        so we can use it in the modulo calculation (modulo needs
        zero-based indexing);

        - add 1 when setting it, because what modulo gives is
        zero-based indexing, and we want to match the pages start
        from 1 condition.

        This gives:

        self.value = ((self.value - 1 + 1) % total_pages) + 1

        ... where -1+1 happens to cancel itself.
        """
        self.app.stdscr.clear()
        self.value = ((self.value - 1 + direction) % self.total_pages()) + 1

    def total_pages(self) -> int:
        return len(range(0, len(self.app.commands[self.app.view]), curses.LINES - 3))

    def get_size(self) -> int:
        return len(self.get_commands())

    def get_commands(self) -> List[str]:
        return self.app.commands[self.app.view][
            (self.value - 1) * (curses.LINES - 3) : self.value * (curses.LINES - 3)
        ]

    def get_selected(self) -> str:
        return self.get_commands()[self.selected.value]


class SelectedCmd:  # pylint: disable=too-few-public-methods
    def __init__(self, page: Page):
        self.value = 0
        self.page = page

    def move(self, direction: int) -> None:
        page_size = self.page.get_size()
        self.value += direction

        try:
            self.value %= page_size
        except ZeroDivisionError:
            return None

        if direction == 1 and self.value == 0:
            self.page.turn(1)
        elif direction == -1 and self.value == (page_size - 1):
            self.page.turn(-1)
            self.value = self.page.get_size() - 1



# pylint: disable=unused-argument
# pylint: disable=unused-import
# pylint: disable=redefined-outer-name
# pylint: disable=too-many-arguments


import os
import shutil
import pytest
from pyhstr import application
from pyhstr.application import App
from pyhstr.utilities import (
    Shell,
    View,
    read,
)

from tests.fixtures import (
    fake_curses,
    fake_stdscr,
    fake_bpython,
    fake_ipython,
    fake_readline,
    fake_standard,
    params,
)


@pytest.mark.python
@pytest.mark.parametrize("command", ["1 + 1 == 2", 'ord("r")', "print(sys.argv)"])
def test_delete_python_history(
    command, fake_stdscr, fake_standard, fake_readline, tmp_path
):
    shutil.copyfile("tests/history/fake_python_history", tmp_path / "history")
    app = App(fake_stdscr)
    app.delete_python_history(command)
    assert command not in read("tests/history/fake_python_history")
    shutil.move(tmp_path / "history", "tests/history/fake_python_history")


@pytest.mark.all
def test_create_search_regex_none(fake_stdscr):
    app = App(fake_stdscr)
    app.regex_mode = True
    app.search_string = "print("
    assert app.create_search_regex() is None


@pytest.mark.all
@pytest.mark.parametrize(
    "search_string, expected, regex_mode, case_sensitivity",
    [
        [
            "spam",
            ['"spam".encode("utf-8")', 'print("SPAM")'],
            False,
            False,
        ],
        [
            "SPAM",
            ['print("SPAM")'],
            False,
            True,
        ],
        [
            "spam",
            ['"spam".encode("utf-8")'],
            False,
            True,
        ],
        [
            "[0-9]+",
            [
                '"spam".encode("utf-8")',
                "[x for x in range(100) if x % 2]",
                "1 + 1 == 2",
                "tau == 2 * pi",
                "4 / 2",
                "2 ** 10",
                "[x for x in range(100)]",
            ],
            True,
            False,
        ],
    ],
)
def test_search(
    search_string,
    expected,
    regex_mode,
    case_sensitivity,
    fake_stdscr,
    fake_curses,
    fake_standard,
):
    app = App(fake_stdscr)
    app.search_string = search_string
    app.regex_mode = regex_mode
    app.case_sensitivity = case_sensitivity
    app.search()
    assert all(x in expected for x in app.commands[app.view])


@pytest.mark.parametrize("shell, fixture", params)
def test_get_history(shell, fixture, fake_stdscr):
    app = App(fake_stdscr)
    assert app.get_history() == read(f"tests/history/fake_{shell.value}_history")


@pytest.mark.all
@pytest.mark.parametrize("regex_mode", [True, False])
def test_toggle_regex_mode(regex_mode, fake_stdscr):
    app = App(fake_stdscr)
    app.regex_mode = regex_mode
    app.toggle_regex_mode()
    assert app.regex_mode == (not regex_mode)


@pytest.mark.all
@pytest.mark.parametrize("case", [True, False])
def test_toggle_case(case, fake_stdscr):
    app = App(fake_stdscr)
    app.case_sensitivity = case
    app.toggle_case()
    assert app.case_sensitivity == (not case)


@pytest.mark.all
@pytest.mark.parametrize(
    "before, expected",
    [
        [View.SORTED, View.FAVORITES],
        [View.FAVORITES, View.ALL],
        [View.ALL, View.SORTED],
    ],
)
def test_toggle_view(before, expected, fake_stdscr):
    app = App(fake_stdscr)
    app.view = before
    app.toggle_view()
    assert app.view == expected


@pytest.mark.parametrize("shell, fixture", params)
def test_add_or_rm_fav(shell, fixture, fake_stdscr):
    app = App(fake_stdscr)
    path = application.SHELLS[shell]["fav"]
    for value in (True, False):
        app.add_or_rm_fav("egg")
        favs = read(path)
        assert favs.__contains__("egg") == value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyhstr', 'tests']

package_data = \
{'': ['*'], 'tests': ['favorites/*', 'history/*']}

setup_kwargs = {
    'name': 'pyhstr',
    'version': '0.5.1',
    'description': 'History suggest box for the standard Python shell, IPython, and bpython',
    'long_description': '# pyhstr\n\n![build status](https://github.com/adder46/pyhstr/workflows/pyhstr/badge.svg) [![codecov](https://codecov.io/gh/adder46/pyhstr/branch/master/graph/badge.svg?token=ZMY0VUX8WS)](https://codecov.io/gh/adder46/pyhstr) [![PyPI version](https://badge.fury.io/py/pyhstr.svg)](https://badge.fury.io/py/pyhstr) ![Python Versions](https://img.shields.io/pypi/pyversions/pyhstr?logo=data:image/svg%2bxml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCAxMDAgMTAwIj4KICA8ZGVmcz4KICAgIDxsaW5lYXJHcmFkaWVudCBpZD0icHlZZWxsb3ciIGdyYWRpZW50VHJhbnNmb3JtPSJyb3RhdGUoNDUpIj4KICAgICAgPHN0b3Agc3RvcC1jb2xvcj0iI2ZlNSIgb2Zmc2V0PSIwLjYiLz4KICAgICAgPHN0b3Agc3RvcC1jb2xvcj0iI2RhMSIgb2Zmc2V0PSIxIi8+CiAgICA8L2xpbmVhckdyYWRpZW50PgogICAgPGxpbmVhckdyYWRpZW50IGlkPSJweUJsdWUiIGdyYWRpZW50VHJhbnNmb3JtPSJyb3RhdGUoNDUpIj4KICAgICAgPHN0b3Agc3RvcC1jb2xvcj0iIzY5ZiIgb2Zmc2V0PSIwLjQiLz4KICAgICAgPHN0b3Agc3RvcC1jb2xvcj0iIzQ2OCIgb2Zmc2V0PSIxIi8+CiAgICA8L2xpbmVhckdyYWRpZW50PgogIDwvZGVmcz4KCiAgPHBhdGggZD0iTTI3LDE2YzAtNyw5LTEzLDI0LTEzYzE1LDAsMjMsNiwyMywxM2wwLDIyYzAsNy01LDEyLTExLDEybC0yNCwwYy04LDAtMTQsNi0xNCwxNWwwLDEwbC05LDBjLTgsMC0xMy05LTEzLTI0YzAtMTQsNS0yMywxMy0yM2wzNSwwbDAtM2wtMjQsMGwwLTlsMCwweiBNODgsNTB2MSIgZmlsbD0idXJsKCNweUJsdWUpIi8+CiAgPHBhdGggZD0iTTc0LDg3YzAsNy04LDEzLTIzLDEzYy0xNSwwLTI0LTYtMjQtMTNsMC0yMmMwLTcsNi0xMiwxMi0xMmwyNCwwYzgsMCwxNC03LDE0LTE1bDAtMTBsOSwwYzcsMCwxMyw5LDEzLDIzYzAsMTUtNiwyNC0xMywyNGwtMzUsMGwwLDNsMjMsMGwwLDlsMCwweiBNMTQwLDUwdjEiIGZpbGw9InVybCgjcHlZZWxsb3cpIi8+CgogIDxjaXJjbGUgcj0iNCIgY3g9IjY0IiBjeT0iODgiIGZpbGw9IiNGRkYiLz4KICA8Y2lyY2xlIHI9IjQiIGN4PSIzNyIgY3k9IjE1IiBmaWxsPSIjRkZGIi8+Cjwvc3ZnPgo=) ![GitHub](https://img.shields.io/github/license/adder46/pyhstr)\n\nInspired by hstr, **pyhstr** is a history suggest box that lets you quickly search, navigate, and manage your Python shell history. At this point, it supports the standard Python shell, IPython, and bpython. The plan is to support ptpython as well, but some help is needed for that to happen (see [issue #7](https://github.com/adder46/pyhstr/issues/7)).\n\n## Installation\n\n\n```\npip install pyhstr\n```\n\n## Usage\n\nThe **standard** shell and **bpython**:\n\n  ```python\n  >>> from pyhstr import hh\n  >>> hh\n  ```\n\n**IPython**:\n\n  ```ipython\n  In [1]: import pyhstr\n  In [2]: %hh\n  ```\n\nMaking an alias should be more convenient though, for example:\n\n```bash\nalias py=\'python3 -ic "from pyhstr import hh"\'\n```\n\n## Screencast\n\n![screenshot](pyhstr.gif)\n\n## Development\n\nYou will need [poetry](https://github.com/python-poetry/poetry), preferably with these options in config:\n\n```toml\nvirtualenvs.create = true\nvirtualenvs.in-project = true\n```\n\nThen clone the repo, cd into it, make a venv, activate it, and install the project:\n\n```sh\ngit clone https://github.com/adder46/pyhstr\ncd pyhstr\npoetry env use python3\n. .venv/bin/activate\npoetry install\n```\n\nTo run tests, mypy checks, and style checks, you need to have Pythons:\n\n- 3.7\n- 3.8\n- 3.9\n\nFor installing all the Python versions, I recommend [pyenv](https://github.com/pyenv/pyenv).\n\nOnce you have them, run:\n\n```\ntox\n```\n\n## Licensing\n\nLicensed under the [MIT License](https://opensource.org/licenses/MIT). For details, see [LICENSE](https://github.com/adder46/pyhstr/blob/master/LICENSE).',
    'author': 'adder46',
    'author_email': 'dedmauz69@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
# pylint: disable=unused-import

import os
import pytest

from pyhstr import utilities
from pyhstr.utilities import (
    Shell,
    detect_shell,
    echo,
    get_bpython_history_path,
    get_ipython_history,
    read,
    remove_duplicates,
    sort,
    write,
)

from tests.fixtures import (
    fake_bpython,
    fake_fcntl,
    fake_ipython,
    fake_standard,
    fake_termios,
    random_history,
    params,
)


@pytest.mark.all
def test_sort():
    history = [3, 2, 4, 6, 2, 4, 3, 3, 4, 5, 6, 3, 2, 4, 5, 5, 3]
    assert sort(history) == [3, 4, 5, 2, 6]


@pytest.mark.all
def test_remove_duplicates():
    history = [3, 2, 4, 6, 2, 4, 3, 3, 4, 5, 6, 3, 2, 4, 5, 5, 3]
    assert remove_duplicates(history) == [3, 2, 4, 6, 5]


@pytest.mark.parametrize("shell, fixture", params)
def test_detect_shell(shell, fixture):
    assert detect_shell() == shell


@pytest.mark.bpython
def test_get_bpython_history_path():
    assert get_bpython_history_path() is not None


@pytest.mark.bpython
def test_get_bpython_history_path_none(monkeypatch):
    monkeypatch.setattr(utilities, "Struct", None)
    assert get_bpython_history_path() is None


@pytest.mark.ipython
@pytest.mark.skipif(os.environ["PYTHON_SHELL"] != "ipython", reason="ipython only")
def test_get_ipython_history(fake_ipython):
    history = get_ipython_history()
    assert isinstance(history, list)


@pytest.mark.all
@pytest.mark.history_length(100)
def test_echo(fake_fcntl, fake_termios, random_history):
    for command in random_history:
        echo(command)
        for byte in command.encode("utf-8"):
            assert (0, None, bytes([byte])) in fake_fcntl.echoed


@pytest.mark.all
@pytest.mark.history_length(100)
def test_write(tmp_path, random_history):
    with pytest.raises(AssertionError):
        write(None, random_history)
    path = tmp_path / "spam"
    write(path, random_history)
    assert (path).exists()


@pytest.mark.all
def test_read_none():
    with pytest.raises(AssertionError):
        read(None)


@pytest.mark.all
def test_read_file_creation(tmp_path):
    assert read(tmp_path / "spam") == [""]


@pytest.mark.all
def test_read_some(tmp_path):
    history = tmp_path / "spam"
    history.write_text("eggs")
    assert read(history) == ["eggs"]
